// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#include "precomp.h"

namespace Dml
{

bool AreAllStridesIdentical(gsl::span<const TensorDesc> tensorDescs)
{
    const size_t tensorDescCount = tensorDescs.size();

    for (size_t i = 1; i < tensorDescCount; ++i)
    {
        gsl::span<const uint32_t> stridesA = tensorDescs[i - 1].GetStrides();
        gsl::span<const uint32_t> stridesB = tensorDescs[i].GetStrides();
        if (stridesA.size() != stridesB.size() || !std::equal(stridesA.begin(), stridesA.end(), stridesB.begin()))
        {
            return false;
        }
    }

    return true;
}

template <typename TOperatorDesc>
class DmlOperatorElementwiseUnary : public DmlOperator
{
public:
    DmlOperatorElementwiseUnary(const MLOperatorKernelCreationContext& kernelInfo) : DmlOperator(kernelInfo)
    {
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetInputCount() == 1);
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetOutputCount() == 1);

        Initialize(kernelInfo, std::nullopt, std::nullopt, kernelInfo.GetTensorShapeDescription().GetOutputTensorShape(0));

        std::vector<DML_TENSOR_DESC> inputDescs = GetDmlInputDescs();
        std::vector<DML_TENSOR_DESC> outputDescs = GetDmlOutputDescs();

        TOperatorDesc opDesc = {};
        opDesc.InputTensor = inputDescs.data();
        opDesc.OutputTensor = outputDescs.data();

        SetDmlOperatorDesc({ ApiTraits::OperatorDescTraits<TOperatorDesc>::Type, &opDesc}, kernelInfo);
    }
};

template<typename T>
void SetFusedActivation(T& opDesc, const DML_OPERATOR_DESC* fusedActivation)
{
    // Activation is only fused for sum operators, which have a template specialization
    THROW_HR(E_INVALIDARG);
}

template<>
void SetFusedActivation(DML_ELEMENT_WISE_ADD1_OPERATOR_DESC& opDesc, const DML_OPERATOR_DESC* fusedActivation)
{                
    opDesc.FusedActivation = fusedActivation;
}

template <typename TOperatorDesc>
class DmlOperatorElementwiseBinary : public DmlOperator
{
public:
    DmlOperatorElementwiseBinary(const MLOperatorKernelCreationContext& kernelInfo) : DmlOperator(kernelInfo)
    {
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetInputCount() == 2);
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetOutputCount() == 1);

        Initialize(kernelInfo, std::nullopt, std::nullopt, kernelInfo.GetTensorShapeDescription().GetOutputTensorShape(0));

        std::vector<DML_TENSOR_DESC> inputDescs = GetDmlInputDescs();
        std::vector<DML_TENSOR_DESC> outputDescs = GetDmlOutputDescs();

        std::optional<ActivationOperatorDesc> fusedActivation = FusionHelpers::TryGetFusedActivationDesc(kernelInfo);
        DML_OPERATOR_DESC fusedActivationDmlDesc = fusedActivation ? fusedActivation->GetDmlDesc() : DML_OPERATOR_DESC();

        TOperatorDesc opDesc = {};
        opDesc.ATensor = &inputDescs[0];
        opDesc.BTensor = &inputDescs[1];
        opDesc.OutputTensor = outputDescs.data();

        DML_OPERATOR_DESC opDescDesc = { ApiTraits::OperatorDescTraits<TOperatorDesc>::Type, &opDesc};

        if (fusedActivation != std::nullopt)
        {    
            // Activation is only fused for two-input sum operators
            THROW_HR_IF(E_INVALIDARG, opDescDesc.Type != DML_OPERATOR_ELEMENT_WISE_ADD1 || kernelInfo.GetInputCount() > 2);

            SetFusedActivation(opDesc, &fusedActivationDmlDesc);
        }

        SetDmlOperatorDesc(opDescDesc, kernelInfo);
    }
};

ComPtr<IDMLCompiledOperator> CreateSecondaryOperator(
    IDMLDevice* dmlDevice,
    DML_EXECUTION_FLAGS executionFlags,
    const DML_OPERATOR_DESC& operatorDesc,
    const MLOperatorKernelCreationContext& kernelInfo
    )
{
    ComPtr<IDMLOperator> dmlOperator;
    ComPtr<IDMLCompiledOperator> compiledOperator;
    THROW_IF_FAILED(dmlDevice->CreateOperator(&operatorDesc, IID_PPV_ARGS(&dmlOperator)));
    THROW_IF_FAILED(dmlDevice->CompileOperator(dmlOperator.Get(), executionFlags, IID_PPV_ARGS(&compiledOperator)));
    return compiledOperator;
}

template <typename TOperatorDesc>
class DmlOperatorElementwiseBinaryLoop : public DmlOperator
{
public:
    DmlOperatorElementwiseBinaryLoop(const MLOperatorKernelCreationContext& kernelInfo) : DmlOperator(kernelInfo)
    {
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetInputCount() >= 1);
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetOutputCount() == 1);

        Initialize(kernelInfo, std::nullopt, std::nullopt, kernelInfo.GetTensorShapeDescription().GetOutputTensorShape(0));

        std::vector<DML_TENSOR_DESC> inputDescs = GetDmlInputDescs();
        std::vector<DML_TENSOR_DESC> outputDescs = GetDmlOutputDescs();
        const size_t inputCount = m_inputTensorDescs.size();

        std::optional<ActivationOperatorDesc> fusedActivation = FusionHelpers::TryGetFusedActivationDesc(kernelInfo);
        DML_OPERATOR_DESC fusedActivationDmlDesc = fusedActivation ? fusedActivation->GetDmlDesc() : DML_OPERATOR_DESC();
        
        // Activation is only fused for two-input sum operators
        THROW_HR_IF(E_INVALIDARG, fusedActivation != std::nullopt && inputCount != 2);

        if (inputCount == 1)
        {
            DML_ELEMENT_WISE_IDENTITY_OPERATOR_DESC identityDesc = {};
            identityDesc.InputTensor = &inputDescs[0];
            identityDesc.OutputTensor = &outputDescs[0];
            SetDmlOperatorDesc({ DML_OPERATOR_ELEMENT_WISE_IDENTITY, &identityDesc }, kernelInfo);
        }
        else
        {
            // Create a single operator that applies to pairwise to every two inputs,
            // accumulated into the output tensor.

            TOperatorDesc opDesc = {};
            opDesc.ATensor = &inputDescs[0];
            opDesc.BTensor = &inputDescs[1];
            opDesc.OutputTensor = outputDescs.data();

            DML_OPERATOR_DESC opDescDesc = { ApiTraits::OperatorDescTraits<TOperatorDesc>::Type, &opDesc};

            if (fusedActivation != std::nullopt)
            {
                SetFusedActivation(opDesc, &fusedActivationDmlDesc);
            }

            SetDmlOperatorDesc(opDescDesc, kernelInfo);

            // If the tensor strides differ between pairs, then it's unsafe to reuse the same operator
            // for all pairs because the wrong stride would be used. So create operators for every additional
            // pair after the first. Given tensors {A, B, C}, the first operator handles A&B, the secondary
            // operator handles tensors B&C, and any additional after that would need another operator.

            if (inputCount >= 2 && !AreAllStridesIdentical(m_inputTensorDescs))
            {
                const DML_EXECUTION_FLAGS executionFlags = GetExecutionFlags();
                gsl::span<const DML_TENSOR_DESC> remainingInputDescs = gsl::make_span(inputDescs);
                remainingInputDescs = remainingInputDescs.subspan(2, remainingInputDescs.size() - 2);

                for (const DML_TENSOR_DESC& tensorDesc : remainingInputDescs)
                {
                    opDesc.ATensor = &tensorDesc;
                    opDesc.BTensor = &outputDescs[0];
                    // Already set - tOpDesc.OutputTensor = &outputDescs[0];
                    m_compiledOperators.push_back(CreateSecondaryOperator(m_dmlDevice.Get(), executionFlags, opDescDesc, kernelInfo));
                }
            }
        }
    }

    void Compute(const MLOperatorKernelContext& kernelContext)
    {
        // For 1 input, just return the input (identity).
        if (m_inputTensorDescs.size() == 1)
        {
            DmlOperator::Compute(kernelContext);
            return;
        }

        // Apply the operator to the first two inputs.
        std::array<IMLOperatorTensor*, 2> inputTensors;
        inputTensors[0] = kernelContext.GetInputTensor(0).GetInterface().Get();
        inputTensors[1] = kernelContext.GetInputTensor(1).GetInterface().Get();

        IMLOperatorTensor* outputTensor = kernelContext.GetOutputTensor(0).GetInterface().Get();
        gsl::span<IMLOperatorTensor*> outputTensors{ &outputTensor, 1 };

        // Combine the first two inputs and store the result in the output tensor.
        THROW_IF_FAILED(m_executionProvider->ExecuteOperator(
            m_compiledOperator.Get(),
            m_persistentResourceBinding ? &*m_persistentResourceBinding : nullptr,
            gsl::make_span(inputTensors),
            outputTensors));

        // For each input after the first two, accumulate into the output tensor.
        for (size_t inputIndex = 2; inputIndex < m_inputTensorDescs.size(); ++inputIndex)
        {
            inputTensors[0] = kernelContext.GetInputTensor(gsl::narrow_cast<uint32_t>(inputIndex)).GetInterface().Get();
            inputTensors[1] = outputTensors[0];

            // Get the next operator for this pair, either reusing the first or using a distinct operator.
            IDMLCompiledOperator* compiledOperator = m_compiledOperators.empty()
                                                   ? m_compiledOperator.Get()
                                                   : m_compiledOperators[inputIndex - 2].Get();

            THROW_IF_FAILED(m_executionProvider->ExecuteOperator(
                compiledOperator,
                m_persistentResourceBinding ? &*m_persistentResourceBinding : nullptr,
                gsl::make_span(inputTensors),
                outputTensors));
        }
    }

    // If multiple compiled operators are needed, beyond m_compiledOperator, they are appended here.
    // The size of the vector will either be empty if all tensor pairs have identical properties,
    // or it will equal inputCount - 2, with the first operator in this vector corresponding to the
    // 3rd input tensor combined with the output of the previous 2 input tensors.
    std::vector<ComPtr<IDMLCompiledOperator>> m_compiledOperators;
};

class DmlOperatorElementwiseMean : public DmlOperator
{
    // Used with 3+ inputs to divide each element by the number of input tensors.
    ComPtr<IDMLCompiledOperator> m_compiledIdentityOp;

public:
    DmlOperatorElementwiseMean(const MLOperatorKernelCreationContext& kernelInfo) : DmlOperator(kernelInfo)
    {
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetInputCount() >= 1);
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetOutputCount() == 1);

        Initialize(kernelInfo, std::nullopt, std::nullopt, kernelInfo.GetTensorShapeDescription().GetOutputTensorShape(0));

        std::vector<DML_TENSOR_DESC> inputDescs = GetDmlInputDescs();
        std::vector<DML_TENSOR_DESC> outputDescs = GetDmlOutputDescs();

        const size_t inputCount = m_inputTensorDescs.size();
        if (inputCount == 1)
        {
            // For 1 input, just return the input
            DML_ELEMENT_WISE_IDENTITY_OPERATOR_DESC identityDesc = {};
            identityDesc.InputTensor = &inputDescs[0];
            identityDesc.OutputTensor = &outputDescs[0];
            SetDmlOperatorDesc({ DML_OPERATOR_ELEMENT_WISE_IDENTITY, &identityDesc }, kernelInfo);
        }
        else if (inputCount == 2)
        {
            // For 2 inputs, use DML's mean operator.
            DML_ELEMENT_WISE_MEAN_OPERATOR_DESC meanDesc = {};
            meanDesc.ATensor = &inputDescs[0];
            meanDesc.BTensor = &inputDescs[1];
            meanDesc.OutputTensor = &outputDescs[0];
            
            SetDmlOperatorDesc({ DML_OPERATOR_ELEMENT_WISE_MEAN, &meanDesc}, kernelInfo);
        }
        else
        {
            // For 3+ inputs, use several DML adds followed by a divide (identity with scale=1/InputCount).
            assert(inputDescs.size() > 2);

            DML_ELEMENT_WISE_ADD_OPERATOR_DESC addDesc = {};
            addDesc.ATensor = &inputDescs[0];
            addDesc.BTensor = &inputDescs[1];
            addDesc.OutputTensor = &outputDescs[0];
            DML_OPERATOR_DESC addDescDesc = { DML_OPERATOR_ELEMENT_WISE_ADD, &addDesc};

            SetDmlOperatorDesc(addDescDesc, kernelInfo);

            if (!AreAllStridesIdentical(m_inputTensorDescs))
            {
                // Create operators for each input after the first two.
                const DML_EXECUTION_FLAGS executionFlags = GetExecutionFlags();
                gsl::span<const DML_TENSOR_DESC> remainingInputDescs = gsl::make_span(inputDescs);
                remainingInputDescs = remainingInputDescs.subspan(2, remainingInputDescs.size() - 2);

                for (const DML_TENSOR_DESC& tensorDesc : remainingInputDescs)
                {
                    addDesc.ATensor = &tensorDesc;
                    addDesc.BTensor = &outputDescs[0];
                    // Already set - addDesc.OutputTensor = &outputDescs[0];
                    m_compiledOperators.push_back(CreateSecondaryOperator(m_dmlDevice.Get(), executionFlags, addDescDesc, kernelInfo));
                }
            }

            // Create division operation using reciprocal of input tensor count.
            DML_SCALE_BIAS scaleBias = {};
            scaleBias.Scale = 1.0f / inputDescs.size();

            DML_ELEMENT_WISE_IDENTITY_OPERATOR_DESC identityDesc = {};
            identityDesc.InputTensor = &outputDescs[0];
            identityDesc.OutputTensor = &outputDescs[0];
            identityDesc.ScaleBias = &scaleBias;

            DML_OPERATOR_DESC identityDescDesc = { DML_OPERATOR_ELEMENT_WISE_IDENTITY, &identityDesc };

            ComPtr<IDMLOperator> identityOp;
            THROW_IF_FAILED(m_dmlDevice->CreateOperator(&identityDescDesc, IID_PPV_ARGS(&identityOp)));

            THROW_IF_FAILED(m_dmlDevice->CompileOperator(identityOp.Get(), GetExecutionFlags(), IID_PPV_ARGS(&m_compiledIdentityOp)));
        }
    }

    void Compute(const MLOperatorKernelContext& kernelContext)
    {
        // Where there's only a single element, just return the input (identity).
        if (m_inputTensorDescs.size() == 1)
        {
            DmlOperator::Compute(kernelContext);
        }
        else if (!m_compiledIdentityOp)
        {
            // Use DML mean operator.
            DmlOperator::Compute(kernelContext);
        }
        else
        {
            // Do N-1 adds followed by a division, where N is the number of inputs.
            std::array<IMLOperatorTensor*, 2> inputTensors;
            inputTensors[0] = kernelContext.GetInputTensor(0).GetInterface().Get();
            inputTensors[1] = kernelContext.GetInputTensor(1).GetInterface().Get();

            IMLOperatorTensor* outputTensor = kernelContext.GetOutputTensor(0).GetInterface().Get();
            gsl::span<IMLOperatorTensor*> outputTensors{ &outputTensor, 1 };

            // Add the first two inputs and store the result in the output tensor.
            THROW_IF_FAILED(m_executionProvider->ExecuteOperator(
                m_compiledOperator.Get(),
                m_persistentResourceBinding ? &*m_persistentResourceBinding : nullptr,
                gsl::make_span(inputTensors),
                outputTensors));

            // For each input after the first two, accumulate into the output tensor.
            for (size_t inputIndex = 2; inputIndex < m_inputTensorDescs.size(); ++inputIndex)
            {
                inputTensors[0] = kernelContext.GetInputTensor(gsl::narrow_cast<uint32_t>(inputIndex)).GetInterface().Get();
                inputTensors[1] = outputTensors[0];

                // Get the next operator for this pair, either reusing the first or using a distinct operator.
                IDMLCompiledOperator* compiledOperator = m_compiledOperators.empty()
                    ? m_compiledOperator.Get()
                    : m_compiledOperators[inputIndex - 2].Get();

                THROW_IF_FAILED(m_executionProvider->ExecuteOperator(
                    compiledOperator,
                    m_persistentResourceBinding ? &*m_persistentResourceBinding : nullptr,
                    gsl::make_span(inputTensors),
                    outputTensors));
            }

            // Dispatch the identity w/ scale operator in-place on the output.
            THROW_IF_FAILED(m_executionProvider->ExecuteOperator(
                m_compiledIdentityOp.Get(),
                nullptr, // persistent resoruce binding
                outputTensors,
                outputTensors));
        }
    }

    // If multiple compiled operators are needed, beyond m_compiledOperator, they are appended here.
    std::vector<ComPtr<IDMLCompiledOperator>> m_compiledOperators;
};

class DmlOperatorElementwiseClip : public DmlOperator
{
public:
    DmlOperatorElementwiseClip(const MLOperatorKernelCreationContext& kernelInfo) : DmlOperator(kernelInfo)
    {
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetInputCount() == 1);
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetOutputCount() == 1);

        Initialize(kernelInfo, std::nullopt, std::nullopt, kernelInfo.GetTensorShapeDescription().GetOutputTensorShape(0));

        std::vector<DML_TENSOR_DESC> inputDescs = GetDmlInputDescs();
        std::vector<DML_TENSOR_DESC> outputDescs = GetDmlOutputDescs();

        DML_ELEMENT_WISE_CLIP_OPERATOR_DESC opDesc = {};
        opDesc.InputTensor = inputDescs.data();
        opDesc.OutputTensor = outputDescs.data();
        opDesc.Min = kernelInfo.GetOptionalAttribute<float>(AttrName::Min, std::numeric_limits<float>::lowest());
        opDesc.Max = kernelInfo.GetOptionalAttribute<float>(AttrName::Max, std::numeric_limits<float>::max());

        SetDmlOperatorDesc({ DML_OPERATOR_ELEMENT_WISE_CLIP, &opDesc}, kernelInfo);
    }
};

class DmlOperatorElementwisePow : public DmlOperator
{
public:
    DmlOperatorElementwisePow(const MLOperatorKernelCreationContext& kernelInfo) : DmlOperator(kernelInfo)
    {
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetInputCount() == 2);
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetOutputCount() == 1);

        Initialize(kernelInfo, std::nullopt, std::nullopt, kernelInfo.GetTensorShapeDescription().GetOutputTensorShape(0));

        std::vector<DML_TENSOR_DESC> inputDescs = GetDmlInputDescs();
        std::vector<DML_TENSOR_DESC> outputDescs = GetDmlOutputDescs();

        DML_ELEMENT_WISE_POW_OPERATOR_DESC opDesc = {};
        opDesc.InputTensor = &inputDescs[0];
        opDesc.ExponentTensor = &inputDescs[1];
        opDesc.OutputTensor = &outputDescs[0];

        SetDmlOperatorDesc({ DML_OPERATOR_ELEMENT_WISE_POW, &opDesc}, kernelInfo);
    }
};

template <typename TOperatorDesc>
class DmlOperatorElementwiseQLinear : public DmlOperator
{
public:
    DmlOperatorElementwiseQLinear(const MLOperatorKernelCreationContext& kernelInfo) : DmlOperator(kernelInfo)
    {
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetInputCount() == 3);
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetOutputCount()  == 1);

        std::vector<uint32_t> outputShape = kernelInfo.GetTensorShapeDescription().GetOutputTensorShape(0);
        const uint32_t outputShapeDimCount = gsl::narrow_cast<uint32_t>(outputShape.size());

        Initialize(kernelInfo, std::nullopt, std::nullopt, outputShape);

        // If the axis attribute is explicitly provided, then broadcasting must be performed along that axis.
        // So massage the actual shapes of the scale and zero-point tensors (1D with length equal to the input
        // axis being broadcast to) into broadcastable shapes.
        if (kernelInfo.HasAttribute(AttrName::Axis, MLOperatorAttributeType::Int))
        {
            const int32_t signedAxis = gsl::narrow_cast<int32_t>(kernelInfo.GetAttribute<int64_t>(AttrName::Axis));
            const uint32_t axis = Dml::HandleNegativeAxis(signedAxis, outputShapeDimCount);
            const uint32_t broadcastAxisLength = outputShape[axis];

            // Explicitly reshape each of the inputs after the first input (scale and zero point tensors).
            for (uint32_t index = 1, inputCount = gsl::narrow_cast<uint32_t>(m_inputTensorDescs.size()); index < inputCount; ++index)
            {
                auto edgeDesc = kernelInfo.GetInputEdgeDescription(index);
                assert(edgeDesc.edgeType == MLOperatorEdgeType::Tensor);

                // Fix up the the tensor shape by filling with trailing ones. So input[2,3] with axis=0 and scale[2]
                // becomes scale[2,1], so that broadcasting works correctly.
                std::vector<uint32_t> adjustedInputTensorShape = kernelInfo.GetTensorShapeDescription().GetInputTensorShape(index);
                ML_CHECK_VALID_ARGUMENT(adjustedInputTensorShape.size() == 1);
                ML_CHECK_VALID_ARGUMENT(adjustedInputTensorShape[0] == broadcastAxisLength);
                adjustedInputTensorShape.insert(adjustedInputTensorShape.end(), outputShapeDimCount - 1 - axis, 1);

                m_inputTensorDescs[index] = TensorDesc(
                    edgeDesc.tensorDataType,
                    gsl::make_span(outputShape),
                    gsl::make_span(adjustedInputTensorShape),
                    TensorAxis::DoNotCoerce,
                    TensorAxis::W,
                    TensorAxis::RightAligned,
                    NchwDimensionCount, // minDimensionCount
                    0 // guaranteedBaseOffsetAlignment
                );
            }
        }

        std::vector<DML_TENSOR_DESC> inputDescs = GetDmlInputDescs();
        std::vector<DML_TENSOR_DESC> outputDescs = GetDmlOutputDescs();

        TOperatorDesc opDesc = {};
        opDesc.InputTensor = &inputDescs[0];
        opDesc.ScaleTensor = &inputDescs[1];
        opDesc.ZeroPointTensor = &inputDescs[2];
        opDesc.OutputTensor = &outputDescs[0];

        SetDmlOperatorDesc({ApiTraits::OperatorDescTraits<TOperatorDesc>::Type, &opDesc}, kernelInfo);
    }
};

class DmlOperatorElementwiseIf : public DmlOperator
{
public:
    DmlOperatorElementwiseIf(const MLOperatorKernelCreationContext& kernelInfo) : DmlOperator(kernelInfo)
    {
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetInputCount() == 3);
        ML_CHECK_VALID_ARGUMENT(kernelInfo.GetOutputCount() == 1);

        Initialize(kernelInfo, std::nullopt, std::nullopt, kernelInfo.GetTensorShapeDescription().GetOutputTensorShape(0));

        std::vector<DML_TENSOR_DESC> inputDescs = GetDmlInputDescs();
        std::vector<DML_TENSOR_DESC> outputDescs = GetDmlOutputDescs();

        DML_ELEMENT_WISE_IF_OPERATOR_DESC opDesc = {};
        opDesc.ConditionTensor = &inputDescs[0];
        opDesc.ATensor = &inputDescs[1];
        opDesc.BTensor = &inputDescs[2];
        opDesc.OutputTensor = &outputDescs[0];

        SetDmlOperatorDesc({ DML_OPERATOR_ELEMENT_WISE_IF, &opDesc }, kernelInfo);
    }
};

// Unary operators:
DML_OP_DEFINE_CREATION_FUNCTION(Sqrt,             DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_SQRT_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Reciprocal,       DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_RECIP_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Cos,              DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_COS_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Sin,              DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_SIN_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Tan,              DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_TAN_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Acos,             DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_ACOS_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Asin,             DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_ASIN_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Atan,             DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_ATAN_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Exp,              DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_EXP_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Log,              DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_LOG_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Abs,              DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_ABS_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Ceil,             DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_CEIL_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Floor,            DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_FLOOR_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Not,              DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_LOGICAL_NOT_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Sign,             DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_SIGN_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(IsNan,            DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_IS_NAN_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Sinh,             DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_SINH_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Cosh,             DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_COSH_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Asinh,            DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_ASINH_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Acosh,            DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_ACOSH_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Atanh,            DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_ATANH_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Erf,              DmlOperatorElementwiseUnary<DML_ELEMENT_WISE_ERF_OPERATOR_DESC>);

// Binary operators:
DML_OP_DEFINE_CREATION_FUNCTION(Greater,          DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_LOGICAL_GREATER_THAN_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Less,             DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_LOGICAL_LESS_THAN_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Equal,            DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_LOGICAL_EQUALS_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(And,              DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_LOGICAL_AND_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Or,               DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_LOGICAL_OR_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Xor,              DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_LOGICAL_XOR_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Add,              DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_ADD_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Sub,              DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_SUBTRACT_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Mul,              DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_MULTIPLY_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Div,              DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_DIVIDE_OPERATOR_DESC>);

// Binary operators that support >2 inputs:
DML_OP_DEFINE_CREATION_FUNCTION(Sum,              DmlOperatorElementwiseBinaryLoop<DML_ELEMENT_WISE_ADD_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Min,              DmlOperatorElementwiseBinaryLoop<DML_ELEMENT_WISE_MIN_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Max,              DmlOperatorElementwiseBinaryLoop<DML_ELEMENT_WISE_MAX_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Mean,             DmlOperatorElementwiseMean);

// Operators with extra attributes:
DML_OP_DEFINE_CREATION_FUNCTION(Clip,             DmlOperatorElementwiseClip);
DML_OP_DEFINE_CREATION_FUNCTION(Pow,              DmlOperatorElementwisePow);
DML_OP_DEFINE_CREATION_FUNCTION(QuantizeLinear,   DmlOperatorElementwiseQLinear<DML_ELEMENT_WISE_QUANTIZE_LINEAR_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(DequantizeLinear, DmlOperatorElementwiseQLinear<DML_ELEMENT_WISE_DEQUANTIZE_LINEAR_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(Where,            DmlOperatorElementwiseIf);

// Fused operators:
DML_OP_DEFINE_CREATION_FUNCTION(FusedAdd,         DmlOperatorElementwiseBinary<DML_ELEMENT_WISE_ADD1_OPERATOR_DESC>);
DML_OP_DEFINE_CREATION_FUNCTION(FusedSum,         DmlOperatorElementwiseBinaryLoop<DML_ELEMENT_WISE_ADD1_OPERATOR_DESC>);

} // namespace Dml
